@@ -1,31 +1,73 @@
 module Data.List.HT.Private where
 
 import Data.List  as List  (find, transpose, unfoldr, isPrefixOf,
-                            findIndices, foldl', )
-import Data.Maybe as Maybe (fromMaybe, catMaybes, )
+                            findIndices, foldl', mapAccumL, )
+import Data.Maybe as Maybe (fromMaybe, catMaybes, isJust, mapMaybe, )
 import Data.Maybe.HT       (toMaybe, )
-import Control.Monad       (guard, msum, )
-import Data.Tuple.HT       (mapPair, mapFst, mapSnd, forcePair, )
+import Control.Monad.HT    ((<=<), )
+import Control.Monad       (guard, msum, mplus, liftM2, )
+import Control.Applicative ((<$>), (<*>), )
+import Data.Tuple.HT       (mapPair, mapFst, mapSnd, forcePair, swap, )
 
-import qualified Data.List.Key.Private   as Key
+import qualified Control.Functor.HT as Func
+
+import qualified Data.List.Comparing     as Cmp
 import qualified Data.List.Match.Private as Match
+import qualified Data.List.Reverse.StrictElement as Rev
 
 import Prelude hiding (unzip, break, span, )
 
 
+{- $setup
+>>> import qualified Test.QuickCheck as QC
+>>> import Test.QuickCheck ((===))
+>>> import Test.Utility (forAllPredicates)
+>>> import Test.QuickCheck (NonNegative(NonNegative), Positive(Positive), NonEmptyList(NonEmpty))
+>>> import qualified Data.List.Match as Match
+>>> import qualified Data.List as List
+>>> import Data.List (transpose)
+>>> import Data.Maybe.HT (toMaybe)
+>>> import Data.Maybe (mapMaybe, isNothing)
+>>> import Data.Char (isLetter, isUpper, toUpper)
+>>> import Data.Eq.HT (equating)
+>>> import Control.Monad (liftM2)
+>>>
+>>> import Prelude hiding (span)
+>>>
+>>> divMaybe :: Int -> Int -> Maybe Int
+>>> divMaybe m n = case divMod n m of (q,0) -> Just q; _ -> Nothing
+>>>
+>>> forAllMaybeFn :: (QC.Testable test) => ((Int -> Maybe Int) -> test) -> QC.Property
+>>> forAllMaybeFn prop = QC.forAll (QC.choose (1,4)) $ prop . divMaybe
+-}
+
+
 -- * Improved standard functions
 
 {- |
 This function is lazier than the one suggested in the Haskell 98 report.
 It is @inits undefined = [] : undefined@,
 in contrast to @Data.List.inits undefined = undefined@.
+
+>>> head $ inits (undefined :: String)
+""
+-}
+{-
+suggested in
+<http://www.haskell.org/pipermail/libraries/2014-July/023291.html>
 -}
 inits :: [a] -> [[a]]
-inits xt =
+inits = map reverse . scanl (flip (:)) []
+
+{- |
+As lazy as 'inits' but less efficient because of repeated 'map'.
+-}
+initsLazy :: [a] -> [[a]]
+initsLazy xt =
    [] :
    case xt of
       [] -> []
-      x:xs -> map (x:) (inits xs)
+      x:xs -> map (x:) (initsLazy xs)
 
 {- |
 Suggested implementation in the Haskell 98 report.
@@ -42,35 +84,61 @@ inits98' =
 
 {- |
 This function is lazier than the one suggested in the Haskell 98 report.
-It is @tails undefined = ([] : undefined) : undefined@,
+It is @tails undefined = undefined : undefined@,
 in contrast to @Data.List.tails undefined = undefined@.
+
+>>> null $ tails (undefined :: String)
+False
 -}
 tails :: [a] -> [[a]]
 tails xt =
-   uncurry (:) $
+   xt :
    case xt of
-      [] -> ([],[])
-      _:xs -> (xt, tails xs)
+      [] -> []
+      _ : xs -> tails xs
 
-tails' :: [a] -> [[a]]
-tails' = fst . breakAfter null . iterate tail
+{- |
+prop> \xs -> tails xs === tailsIterate (xs :: [Int])
+-}
+tailsIterate :: [a] -> [[a]]
+tailsIterate = fst . breakAfter null . iterate tail
 
 tails98            :: [a] -> [[a]]
 tails98 []         = [[]]
 tails98 xxs@(_:xs) = xxs : tails98 xs
 
+
+type NonEmpty a = (a, [a])
+
+{- |
+For custom @NonEmpty@ type, see @non-empty@ package, or @base>=4.9.0.0@.
+
+>>> nonEmptyTails [1,2,3::Int]
+[(1,[2,3]),(2,[3]),(3,[])]
+
+prop> \xs -> tails xs === map (uncurry (:)) (nonEmptyTails xs) ++ [[]::[Int]]
+-}
+nonEmptyTails :: [a] -> [NonEmpty a]
+nonEmptyTails =
+   let go [] = []
+       go (a:as) = (a,as) : go as
+   in go
+
+
 {- |
 This function compares adjacent elements of a list.
 If two adjacent elements satisfy a relation then they are put into the same sublist.
 Example:
 
-> groupBy (<) "abcdebcdef"  ==  ["abcde","bcdef"]
+>>> groupBy (<) "abcdebcdef"
+["abcde","bcdef"]
 
 In contrast to that 'Data.List.groupBy' compares
 the head of each sublist with each candidate for this sublist.
 This yields
 
-> List.groupBy (<) "abcdebcdef"  ==  ["abcdebcdef"]
+>>> List.groupBy (<) "abcdebcdef"
+["abcdebcdef"]
 
 The second @'b'@ is compared with the leading @'a'@.
 Thus it is put into the same sublist as @'a'@.
@@ -79,8 +147,15 @@ The sublists are never empty.
 Thus the more precise result type would be @[(a,[a])]@.
 -}
 groupBy :: (a -> a -> Bool) -> [a] -> [[a]]
-groupBy = Key.groupBy
+groupBy = Cmp.group
 
+{- |
+>>> head $ group ("aabcdef"++repeat 'g')
+"aa"
+
+>>> take 10 $ group ("aabcdef"++cycle "gh")
+["aa","b","c","d","e","f","g","h","g","h"]
+-}
 group :: (Eq a) => [a] -> [[a]]
 group = groupBy (==)
 
@@ -117,16 +192,30 @@ partition p =
 {- |
 It is @Data.List.span f undefined = undefined@,
 whereas @span f undefined = (undefined, undefined)@.
+
+>>> case span (>0) (undefined :: [Int]) of (_,_) -> "defined"
+"defined"
 -}
-span, break :: (a -> Bool) -> [a] -> ([a],[a])
+span, spanCase, break :: (a -> Bool) -> [a] -> ([a],[a])
 span p =
    let recourse xt =
-          forcePair $
           fromMaybe ([],xt) $
           do (x,xs) <- viewL xt
              guard $ p x
              return $ mapFst (x:) $ recourse xs
-   in  recourse
+   in forcePair . recourse
+
+{- |
+prop> \xs -> span (>0) xs === spanCase (>0) (xs::[Int])
+-}
+spanCase p =
+   let recourse [] = ([], [])
+       recourse xt@(x : xs) =
+         if p x
+         then mapFst (x:) $ recourse xs
+         else ([], xt)
+   in forcePair . recourse
+
 
 break p =  span (not . p)
 
@@ -140,21 +229,20 @@ Remove the separators.
 This is somehow a generalization of 'lines' and 'words'.
 But note the differences:
 
-> Prelude Data.List.HT> words "a  a"
-> ["a","a"]
-> Prelude Data.List.HT> chop (' '==) "a  a"
-> ["a","","a"]
+>>> words "a  a"
+["a","a"]
+>>> chop (' '==) "a  a"
+["a","","a"]
 
-> Prelude Data.List.HT> lines "a\n\na"
-> ["a","","a"]
-> Prelude Data.List.HT> chop ('\n'==) "a\n\na"
-> ["a","","a"]
-
-> Prelude Data.List.HT> lines "a\n"
-> ["a"]
-> Prelude Data.List.HT> chop ('\n'==) "a\n"
-> ["a",""]
+>>> lines "a\n\na"
+["a","","a"]
+>>> chop ('\n'==) "a\n\na"
+["a","","a"]
 
+>>> lines "a\n"
+["a"]
+>>> chop ('\n'==) "a\n"
+["a",""]
 -}
 chop :: (a -> Bool) -> [a] -> [[a]]
 chop p =
@@ -170,7 +258,7 @@ chop' p =
    in  recourse
 
 
-chopAtRun :: (Eq a) => (a -> Bool) -> [a] -> [[a]]
+chopAtRun :: (a -> Bool) -> [a] -> [[a]]
 chopAtRun p =
    let recourse [] = [[]]
        recourse y =
@@ -179,11 +267,52 @@ chopAtRun p =
    in  recourse
 
 
+{- |
+>>> chopNothing ([] :: [Maybe String])
+[[]]
+
+>>> chopNothing [Nothing :: Maybe String]
+[[],[]]
+
+>>> chopNothing [Just "a", Just "b", Nothing, Just "c", Nothing, Just "d", Just "e"]
+[["a","b"],["c"],["d","e"]]
+
+prop> \xs ys -> chopNothing (xs ++ Nothing : ys) === chopNothing xs ++ chopNothing (ys :: [Maybe Int])
+
+prop> \ns xs -> equating (Match.take (ns::[()])) (cycle (chopNothing xs)) (chopNothing (cycle (xs ++ [Nothing :: Maybe Int])))
+
+
+Inspired by
+<https://stackoverflow.com/questions/29095967/splitting-list-into-sublists-along-elements>.
+-}
+chopNothing :: [Maybe a] -> [[a]]
+chopNothing = chopNothingCase
+
+chopNothingCase :: [Maybe a] -> [[a]]
+chopNothingCase =
+   let go xs =
+         case spanJust id xs of
+            (ys,zs) -> ys : case zs of [] -> []; _:ws -> go ws
+   in go
+
+chopNothingSwitchL :: [Maybe a] -> [[a]]
+chopNothingSwitchL =
+   let go xs =
+         case spanJust id xs of
+            (ys,zs) -> ys : switchL [] (const go) zs
+   in go
+
+
 {- |
 Like 'break', but splits after the matching element.
+
+prop> forAllPredicates $ \p xs -> uncurry (++) (breakAfter p xs) === xs
 -}
 breakAfter :: (a -> Bool) -> [a] -> ([a], [a])
-breakAfter p =
+breakAfter = breakAfterRec
+
+breakAfterRec :: (a -> Bool) -> [a] -> ([a], [a])
+breakAfterRec p =
    let recourse [] = ([],[])
        recourse (x:xs) =
           mapFst (x:) $
@@ -192,16 +321,62 @@ breakAfter p =
             else recourse xs
    in  forcePair . recourse
 
+{-
+The use of 'foldr' might allow for fusion,
+but unfortunately this simple implementation would copy the tail of the list.
+-}
+-- | prop> forAllPredicates $ \p xs -> breakAfterRec p xs === breakAfterFoldr p xs
+breakAfterFoldr :: (a -> Bool) -> [a] -> ([a], [a])
+breakAfterFoldr p =
+   forcePair .
+   foldr
+      (\x yzs -> mapFst (x:) $ if p x then ([], uncurry (++) yzs) else yzs)
+      ([],[])
+
+-- | prop> forAllPredicates $ \p xs -> breakAfterRec p xs === breakAfterBreak p xs
+breakAfterBreak :: (a -> Bool) -> [a] -> ([a], [a])
+breakAfterBreak p xs =
+   case break p xs of
+      (ys, []) -> (ys, [])
+      (ys, z:zs) -> (ys++[z], zs)
+
+-- | prop> forAllPredicates $ \p xs -> breakAfterRec p xs === breakAfterTakeUntil p xs
+breakAfterTakeUntil :: (a -> Bool) -> [a] -> ([a], [a])
+breakAfterTakeUntil p xs =
+   forcePair $
+   (\ys -> (map fst ys, maybe [] (snd . snd) $ viewR ys)) $
+   takeUntil (p . fst) $ zip xs $ tail $ tails xs
+
+{- |
+Take all elements until one matches.
+The matching element is returned, too.
+This is the key difference to @takeWhile (not . p)@.
+It holds:
+
+prop> forAllPredicates $ \p xs -> takeUntil p xs === fst (breakAfter p xs)
+-}
+takeUntil :: (a -> Bool) -> [a] -> [a]
+takeUntil p = foldr (\x ys -> x : if p x then [] else ys) []
+
 
 {- |
 Split the list after each occurence of a terminator.
 Keep the terminator.
 There is always a list for the part after the last terminator.
 It may be empty.
+See package @non-empty@ for more precise result type.
+
+prop> forAllPredicates $ \p xs -> concat (segmentAfter p xs) === xs
+prop> forAllPredicates $ \p xs -> length (filter p xs) === length (tail (segmentAfter p xs))
+prop> forAllPredicates $ \p -> all (p . last) . init . segmentAfter p
+prop> forAllPredicates $ \p -> all (all (not . p) . init) . init . segmentAfter p
+
+This test captures both infinitely many groups and infinitely big groups:
+
+prop> forAllPredicates $ \p x -> flip seq True . (!!100) . concat . segmentAfter p . cycle . (x:)
 -}
 segmentAfter :: (a -> Bool) -> [a] -> [[a]]
 segmentAfter p =
---   foldr (\ x ~yt@(y:ys) -> if p x then [x]:yt else (x:y):ys) [[]]
    uncurry (:) .
    foldr
       (\x ~(y,ys) ->
@@ -209,34 +384,27 @@ segmentAfter p =
          if p x then ([],y:ys) else (y,ys))
       ([],[])
 
-propSegmentAfterConcat :: Eq a => (a -> Bool) -> [a] -> Bool
-propSegmentAfterConcat p xs =
-   concat (segmentAfter p xs) == xs
-
-propSegmentAfterNumSeps :: (a -> Bool) -> [a] -> Bool
-propSegmentAfterNumSeps p xs =
-   length (filter p xs) == length (tail (segmentAfter p xs))
-
-propSegmentAfterLasts :: (a -> Bool) -> [a] -> Bool
-propSegmentAfterLasts p =
-   all (p . last) . init . segmentAfter p
-
-propSegmentAfterInits :: (a -> Bool) -> [a] -> Bool
-propSegmentAfterInits p =
-   all (all (not . p) . init) . init . segmentAfter p
-
-{-
-This test captures both infinitely many groups and infinitely big groups.
--}
-propSegmentAfterInfinite :: (a -> Bool) -> a -> [a] -> Bool
-propSegmentAfterInfinite p x =
-   flip seq True . (!!100) . concat . segmentAfter p . cycle . (x:)
+segmentAfter' :: (a -> Bool) -> [a] -> [[a]]
+segmentAfter' p =
+   foldr (\ x ~yt@(y:ys) -> if p x then [x]:yt else (x:y):ys) [[]]
 
 {- |
 Split the list before each occurence of a leading character.
 Keep these characters.
 There is always a list for the part before the first leading character.
 It may be empty.
+See package @non-empty@ for more precise result type.
+
+>>> segmentBefore isUpper "AbcdXyz"
+["","Abcd","Xyz"]
+>>> segmentBefore isUpper "kAbcdXYZ"
+["k","Abcd","X","Y","Z"]
+
+prop> forAllPredicates $ \p xs -> concat (segmentBefore p xs) === xs
+prop> forAllPredicates $ \p xs -> length (filter p xs) === length (tail (segmentBefore p xs))
+prop> forAllPredicates $ \p -> all (p . head) . tail . segmentBefore p
+prop> forAllPredicates $ \p -> all (all (not . p) . tail) . tail . segmentBefore p
+prop> forAllPredicates $ \p x -> flip seq True . (!!100) . concat . segmentBefore p . cycle . (x:)
 -}
 segmentBefore :: (a -> Bool) -> [a] -> [[a]]
 segmentBefore p =
@@ -248,6 +416,7 @@ segmentBefore p =
          in  if p x then ([],xs:ys) else (xs,ys))
       ([],[])
 
+-- | prop> forAllPredicates $ \p xs -> segmentBefore p xs === segmentBefore' p xs
 segmentBefore' :: (a -> Bool) -> [a] -> [[a]]
 segmentBefore' p =
    uncurry (:) .
@@ -258,6 +427,7 @@ segmentBefore' p =
          return (x:xs, xss)) .
    groupBy (\_ x -> not $ p x)
 
+-- | prop> forAllPredicates $ \p xs -> segmentBefore p xs === segmentBefore'' p xs
 segmentBefore'' :: (a -> Bool) -> [a] -> [[a]]
 segmentBefore'' p =
    (\xst ->
@@ -268,33 +438,64 @@ segmentBefore'' p =
    (error "segmentBefore: dummy element" :)
 
 
-propSegmentBeforeConcat :: Eq a => (a -> Bool) -> [a] -> Bool
-propSegmentBeforeConcat p xs =
-   concat (segmentBefore p xs) == xs
+{- |
+>>> segmentBeforeJust (\c -> toMaybe (isLetter c) (toUpper c)) "123a5345b---"
+("123",[('A',"5345"),('B',"---")])
+-}
+segmentBeforeJust ::
+   (a -> Maybe b) ->
+   [a] -> ([a], [(b, [a])])
+segmentBeforeJust f =
+   forcePair .
+   foldr
+      (\ x ~(y,ys) ->
+         case f x of
+            Just b -> ([],(b,y):ys)
+            Nothing -> (x:y,ys))
+      ([],[])
 
-propSegmentBeforeNumSeps :: (a -> Bool) -> [a] -> Bool
-propSegmentBeforeNumSeps p xs =
-   length (filter p xs) == length (tail (segmentBefore p xs))
+{- |
+>>> segmentAfterJust (\c -> toMaybe (isLetter c) (toUpper c)) "123a5345b---"
+([("123",'A'),("5345",'B')],"---")
+-}
+segmentAfterJust ::
+   (a -> Maybe b) ->
+   [a] -> ([([a], b)], [a])
+segmentAfterJust f =
+   swap .
+   uncurry (mapAccumL (\as0 (b,as1) -> (as1, (as0,b)))) .
+   segmentBeforeJust f
 
-propSegmentBeforeHeads :: (a -> Bool) -> [a] -> Bool
-propSegmentBeforeHeads p =
-   all (p . head) . tail . segmentBefore p
 
-propSegmentBeforeTails :: (a -> Bool) -> [a] -> Bool
-propSegmentBeforeTails p =
-   all (all (not . p) . tail) . tail . segmentBefore p
+{- |
+>>> segmentBeforeRight [Left 'a', Right LT, Right GT, Left 'b']
+("a",[(LT,""),(GT,"b")])
 
-propSegmentBeforeInfinite :: (a -> Bool) -> a -> [a] -> Bool
-propSegmentBeforeInfinite p x =
-   flip seq True . (!!100) . concat . segmentBefore p . cycle . (x:)
+prop> forAllMaybeFn $ \f xs -> segmentBeforeJust f xs === segmentBeforeRight (map (\x -> maybe (Left x) Right (f x)) xs)
+-}
+segmentBeforeRight ::
+   [Either a b] -> ([a], [(b, [a])])
+segmentBeforeRight =
+   forcePair .
+   foldr
+      (\ x ~(y,ys) ->
+         case x of
+            Right b -> ([],(b,y):ys)
+            Left a -> (a:y,ys))
+      ([],[])
 
-propSegmentBeforeGroupBy0 :: Eq a => (a -> Bool) -> [a] -> Bool
-propSegmentBeforeGroupBy0 p xs =
-   segmentBefore p xs == segmentBefore' p xs
+{- |
+>>> segmentAfterRight [Left 'a', Right LT, Right GT, Left 'b']
+([("a",LT),("",GT)],"b")
 
-propSegmentBeforeGroupBy1 :: Eq a => (a -> Bool) -> [a] -> Bool
-propSegmentBeforeGroupBy1 p xs =
-   segmentBefore p xs == segmentBefore'' p xs
+prop> forAllMaybeFn $ \f xs -> segmentAfterJust f xs === segmentAfterRight (map (\x -> maybe (Left x) Right (f x)) xs)
+-}
+segmentAfterRight ::
+   [Either a b] -> ([([a], b)], [a])
+segmentAfterRight =
+   swap .
+   uncurry (mapAccumL (\as0 (b,as1) -> (as1, (as0,b)))) .
+   segmentBeforeRight
 
 
 -- cf. Matroid.hs
@@ -309,19 +510,30 @@ and thus can also be used for lists of functions.
 
 See also the proposal
  <http://www.haskell.org/pipermail/libraries/2008-February/009270.html>
+
+>>> removeEach "abc"
+[('a',"bc"),('b',"ac"),('c',"ab")]
+>>> removeEach "a"
+[('a',"")]
+>>> removeEach ""
+[]
 -}
 removeEach :: [a] -> [(a, [a])]
 removeEach =
    map (\(ys, pivot, zs) -> (pivot,ys++zs)) . splitEverywhere
 
+{- |
+>>> splitEverywhere "abc"
+[("",'a',"bc"),("a",'b',"c"),("ab",'c',"")]
+>>> splitEverywhere "a"
+[("",'a',"")]
+>>> splitEverywhere ""
+[]
+-}
+{-# INLINE splitEverywhere #-}
 splitEverywhere :: [a] -> [([a], a, [a])]
 splitEverywhere xs =
-   map
-      (\(y, zs0) ->
-         case zs0 of
-            z:zs -> (y,z,zs)
-            [] -> error "splitEverywhere: empty list")
-      (init (zip (inits xs) (tails xs)))
+   zipWith (\ys (z,zs) -> (ys,z,zs)) (inits xs) (nonEmptyTails xs)
 
 
 
@@ -333,6 +545,8 @@ It holds @splitLast xs == (init xs, last xs)@,
 but 'splitLast' is more efficient
 if the last element is accessed after the initial ones,
 because it avoids memoizing list.
+
+prop> \(NonEmpty xs) -> splitLast (xs::String)  ===  (init xs, last xs)
 -}
 splitLast :: [a] -> ([a], a)
 splitLast [] = error "splitLast: empty list"
@@ -340,10 +554,6 @@ splitLast [x] = ([], x)
 splitLast (x:xs) =
    let (xs', lastx) = splitLast xs in (x:xs', lastx)
 
-propSplitLast :: Eq a => [a] -> Bool
-propSplitLast xs =
-   splitLast xs  ==  (init xs, last xs)
-
 
 {- |
 Should be prefered to 'head' and 'tail'.
@@ -355,17 +565,13 @@ viewL []     = Nothing
 
 {- |
 Should be prefered to 'init' and 'last'.
+
+prop> \xs -> maybe True ((init xs, last xs) == ) (viewR (xs::String))
 -}
 viewR :: [a] -> Maybe ([a], a)
 viewR =
    foldr (\x -> Just . forcePair . maybe ([],x) (mapFst (x:))) Nothing
 
-propViewR :: Eq a => [a] -> Bool
-propViewR xs =
-   maybe True
-      ((init xs, last xs) == )
-      (viewR xs)
-
 {- |
 Should be prefered to 'head' and 'tail'.
 -}
@@ -380,57 +586,43 @@ switchL' n j =
 
 {- |
 Should be prefered to 'init' and 'last'.
+
+prop> \xs -> switchR True (\ixs lxs -> ixs == init xs && lxs == last xs) (xs::String)
 -}
 {-# INLINE switchR #-}
 switchR :: b -> ([a] -> a -> b) -> [a] -> b
 switchR n j =
    maybe n (uncurry j) . viewR
 
-propSwitchR :: Eq a => [a] -> Bool
-propSwitchR xs =
-   switchR True (\ixs lxs -> ixs == init xs && lxs == last xs) xs
-
-
 
 -- * List processing starting at the end
 
 {- |
-Remove the longest suffix of elements satisfying p.
-In contrast to @reverse . dropWhile p . reverse@
-this works for infinite lists, too.
--}
-dropWhileRev :: (a -> Bool) -> [a] -> [a]
-dropWhileRev p =
-   foldr (\x xs -> if p x && null xs then [] else x:xs) []
+@takeRev n@ is like @reverse . take n . reverse@
+but it is lazy enough to work for infinite lists, too.
 
-dropWhileRev' :: (a -> Bool) -> [a] -> [a]
-dropWhileRev' p =
-   concat . init . segmentAfter (not . p)
+prop> \n xs -> takeRev n (xs::String) === reverse (take n (reverse xs))
+-}
+takeRev :: Int -> [a] -> [a]
+takeRev n xs = Match.drop (drop n xs) xs
 
 {- |
-Alternative version of @reverse . takeWhile p . reverse@.
+@dropRev n@ is like @reverse . drop n . reverse@
+but it is lazy enough to work for infinite lists, too.
+
+prop> \n xs -> dropRev n (xs::String) === reverse (drop n (reverse xs))
 -}
-takeWhileRev :: (a -> Bool) -> [a] -> [a]
-takeWhileRev p =
-   last . segmentAfter (not . p)
+dropRev :: Int -> [a] -> [a]
+dropRev n xs = Match.take (drop n xs) xs
 
 {- |
-Doesn't seem to be superior to the naive implementation.
--}
-takeWhileRev' :: (a -> Bool) -> [a] -> [a]
-takeWhileRev' p =
-   (\xs -> if fst (head xs)
-             then map snd xs
-             else []) .
-   last . Key.aux groupBy (==) p
+@splitAtRev n xs == (dropRev n xs, takeRev n xs)@.
 
-{- |
-However it is more inefficient,
-because of repeatedly appending single elements. :-(
+prop> \n xs -> splitAtRev n (xs::String) === (dropRev n xs, takeRev n xs)
+prop> \n xs -> (xs::String) === uncurry (++) (splitAtRev n xs)
 -}
-takeWhileRev'' :: (a -> Bool) -> [a] -> [a]
-takeWhileRev'' p =
-   foldl (\xs x -> if p x then xs++[x] else []) []
+splitAtRev :: Int -> [a] -> ([a], [a])
+splitAtRev n xs = Match.splitAt (drop n xs) xs
 
 
 -- * List processing with Maybe and Either
@@ -439,17 +631,41 @@ takeWhileRev'' p =
 @maybePrefixOf xs ys@ is @Just zs@ if @xs@ is a prefix of @ys@,
 where @zs@ is @ys@ without the prefix @xs@.
 Otherwise it is @Nothing@.
+It is the same as 'Data.List.stripPrefix'.
+
+>>> maybePrefixOf "abc" "abcdef"
+Just "def"
+>>> maybePrefixOf "def" "abcdef"
+Nothing
 -}
 maybePrefixOf :: Eq a => [a] -> [a] -> Maybe [a]
 maybePrefixOf (x:xs) (y:ys) = guard (x==y) >> maybePrefixOf xs ys
 maybePrefixOf [] ys = Just ys
 maybePrefixOf _  [] = Nothing
 
+{- |
+>>> maybeSuffixOf "abc" "abcdef"
+Nothing
+>>> maybeSuffixOf "def" "abcdef"
+Just "abc"
+
+prop> \xs ys -> maybeSuffixOf xs ys === fmap reverse (maybePrefixOf (reverse xs) (reverse (ys::[Ordering])))
+-}
+maybeSuffixOf :: Eq a => [a] -> [a] -> Maybe [a]
+maybeSuffixOf [] bs = Just bs -- Avoid copying in the trivial case
+maybeSuffixOf as bs = do
+   prefixLen <- Match.dropMaybe as bs
+   let (prefix, suffix) = Match.splitAt prefixLen bs
+   guard (as == suffix)
+   return prefix
+
+
+
 {- |
 Partition a list into elements which evaluate to @Just@ or @Nothing@ by @f@.
 
-It holds @mapMaybe f == fst . partitionMaybe f@
-and @partition p == partitionMaybe (\ x -> toMaybe (p x) x)@.
+prop> forAllMaybeFn $ \f xs -> partitionMaybe f xs === (mapMaybe f xs, filter (isNothing . f) xs)
+prop> forAllPredicates $ \p xs -> partition p xs === partitionMaybe (\x -> toMaybe (p x) x) xs
 -}
 partitionMaybe :: (a -> Maybe b) -> [a] -> ([b], [a])
 partitionMaybe f =
@@ -462,14 +678,76 @@ partitionMaybe f =
 This is the cousin of 'takeWhile'
 analogously to 'catMaybes' being the cousin of 'filter'.
 
+>>> takeWhileJust [Just 'a', Just 'b', Nothing, Just 'c']
+"ab"
+
 Example: Keep the heads of sublists until an empty list occurs.
 
-> takeWhileJust $ map (fmap fst . viewL) xs
+>>> takeWhileJust $ map (fmap fst . viewL) ["abc","def","","xyz"]
+"ad"
+
+For consistency with 'takeWhile',
+'partitionMaybe' and 'dropWhileNothing' it should have been:
+
+> takeWhileJust_ :: (a -> Maybe b) -> a -> [b]
+
+However, both variants are interchangeable:
+
+> takeWhileJust_ f === takeWhileJust . map f
+> takeWhileJust === takeWhileJust_ id
 -}
 takeWhileJust :: [Maybe a] -> [a]
 takeWhileJust =
    foldr (\x acc -> maybe [] (:acc) x) []
 
+dropWhileNothing :: (a -> Maybe b) -> [a] -> Maybe (b, [a])
+dropWhileNothing f =
+   msum . map (Func.mapFst f <=< viewL) . tails
+
+-- | prop> forAllMaybeFn $ \f xs -> dropWhileNothing f xs === dropWhileNothingRec f xs
+dropWhileNothingRec :: (a -> Maybe b) -> [a] -> Maybe (b, [a])
+dropWhileNothingRec f =
+   let go [] = Nothing
+       go (a:xs) = (flip (,) xs <$> f a) `mplus` go xs
+   in  go
+
+-- | prop> forAllMaybeFn $ \f xs -> snd (breakJust f xs) === dropWhileNothing f xs
+breakJust :: (a -> Maybe b) -> [a] -> ([a], Maybe (b, [a]))
+breakJust f =
+   let go [] = ([], Nothing)
+       go (a:xs) =
+         case f a of
+            Nothing -> mapFst (a:) $ go xs
+            Just b -> ([], Just (b, xs))
+   in  go
+
+-- memory leak, because xs is hold all the time
+-- | prop> forAllMaybeFn $ \f xs -> breakJust f xs === breakJustRemoveEach f xs
+breakJustRemoveEach :: (a -> Maybe b) -> [a] -> ([a], Maybe (b, [a]))
+breakJustRemoveEach f xs =
+   switchL (xs, Nothing) const $
+   mapMaybe (\(ys,a,zs) -> (\b -> (ys, Just (b,zs))) <$> f a) $
+   splitEverywhere xs
+
+-- needs to apply 'f' twice at the end and uses partial functions
+-- | prop> forAllMaybeFn $ \f xs -> breakJust f xs === breakJustPartial f xs
+breakJustPartial :: (a -> Maybe b) -> [a] -> ([a], Maybe (b, [a]))
+breakJustPartial f xs =
+   let (ys,zs) = break (isJust . f) xs
+   in  (ys,
+        mapFst (maybe (error "breakJust: unexpected Nothing") id . f) <$>
+            viewL zs)
+
+spanJust :: (a -> Maybe b) -> [a] -> ([b], [a])
+spanJust f =
+   let go [] = ([], [])
+       go xt@(a:xs) =
+         case f a of
+            Just b -> mapFst (b:) $ go xs
+            Nothing -> ([], xt)
+   in  go
+
+
 unzipEithers :: [Either a b] -> ([a], [b])
 unzipEithers =
    forcePair .
@@ -478,65 +756,72 @@ unzipEithers =
 
 -- * Sieve and slice
 
-{-| keep every k-th value from the list -}
+{- | keep every k-th value from the list
+
+>>> sieve 6 ['a'..'z']
+"agmsy"
+-}
 sieve, sieve', sieve'', sieve''' :: Int -> [a] -> [a]
 sieve k =
    unfoldr (\xs -> toMaybe (not (null xs)) (head xs, drop k xs))
 
+-- | prop> \(Positive n) xs -> sieve n xs === sieve' n (xs::String)
 sieve' k = map head . sliceVertical k
 
+-- | prop> \(Positive n) xs -> sieve n xs === sieve'' n (xs::String)
 sieve'' k x = map (x!!) [0,k..(length x-1)]
 
+-- | prop> \(Positive n) xs -> sieve n xs === sieve''' n (xs::String)
 sieve''' k = map head . takeWhile (not . null) . iterate (drop k)
 
-propSieve :: Eq a => Int -> [a] -> Bool
-propSieve n x =
-   sieve n x == sieve'  n x   &&
-   sieve n x == sieve'' n x
 
 {-
 sliceHorizontal is faster than sliceHorizontal' but consumes slightly more memory
 (although it needs no swapping)
 -}
+{- |
+>>> sliceHorizontal 6 ['a'..'z']
+["agmsy","bhntz","ciou","djpv","ekqw","flrx"]
+
+prop> \(NonEmpty xs) -> QC.forAll (QC.choose (1, length xs)) $ \n -> sliceHorizontal n xs === transpose (sliceVertical n (xs::String))
+prop> \(NonEmpty xs) -> QC.forAll (QC.choose (1, length xs)) $ \n -> sliceVertical  n xs === transpose (sliceHorizontal n (xs::String))
+
+The properties do not hold for empty lists because of:
+
+>>> sliceHorizontal 4 ([]::[Int])
+[[],[],[],[]]
+-}
 sliceHorizontal, sliceHorizontal', sliceHorizontal'', sliceHorizontal''' ::
    Int -> [a] -> [[a]]
 sliceHorizontal n =
    map (sieve n) . take n . iterate (drop 1)
 
+-- | prop> \(NonNegative n) xs -> sliceHorizontal n xs === sliceHorizontal' n (xs::String)
 sliceHorizontal' n =
    foldr (\x ys -> let y = last ys in Match.take ys ((x:y):ys)) (replicate n [])
 
+-- | prop> \(Positive n) xs -> sliceHorizontal n xs === sliceHorizontal'' n (xs::String)
 sliceHorizontal'' n =
    reverse . foldr (\x ~(y:ys) -> ys ++ [x:y]) (replicate n [])
 
 sliceHorizontal''' n =
    take n . transpose . takeWhile (not . null) . iterate (drop n)
 
-propSliceHorizontal :: Eq a => Int -> [a] -> Bool
-propSliceHorizontal n x =
-   sliceHorizontal n x == sliceHorizontal'   n x &&
-   sliceHorizontal n x == sliceHorizontal''  n x &&
-   sliceHorizontal n x == sliceHorizontal''' n x
-
 
+{- |
+>>> sliceVertical 6 ['a'..'z']
+["abcdef","ghijkl","mnopqr","stuvwx","yz"]
+-}
 sliceVertical, sliceVertical' :: Int -> [a] -> [[a]]
 sliceVertical n =
    map (take n) . takeWhile (not . null) . iterate (drop n)
       {- takeWhile must be performed before (map take)
          in order to handle (n==0) correctly -}
 
+-- | prop> \(NonNegative n) xs -> equating (take 100000) (sliceVertical n xs) (sliceVertical' n (xs::String))
 sliceVertical' n =
    unfoldr (\x -> toMaybe (not (null x)) (splitAt n x))
 
-propSliceVertical :: Eq a => Int -> [a] -> Bool
-propSliceVertical n x =
-   take 100000 (sliceVertical n x) == take 100000 (sliceVertical' n x)
-
-propSlice :: Eq a => Int -> [a] -> Bool
-propSlice n x =
-   -- problems: sliceHorizontal 4 [] == [[],[],[],[]]
-   sliceHorizontal n x == transpose (sliceVertical  n x)  &&
-   sliceVertical  n x == transpose (sliceHorizontal n x)
 
 
 
@@ -545,6 +830,10 @@ propSlice n x =
 search :: (Eq a) => [a] -> [a] -> [Int]
 search sub str = findIndices (isPrefixOf sub) (tails str)
 
+{- |
+prop> \(NonEmpty xs) ys -> replace xs xs ys === (ys::String)
+prop> \(NonEmpty xs) (NonEmpty ys) -> equating (take 1000) (replace xs ys (cycle xs)) (cycle (ys::String))
+-}
 replace :: Eq a => [a] -> [a] -> [a] -> [a]
 replace src dst =
    let recourse [] = []
@@ -569,14 +858,6 @@ replace' :: (Eq a) => [a] -> [a] -> [a] -> [a]
 replace' src dst xs =
    concatMap (fromMaybe dst) (markSublists src xs)
 
-propReplaceId :: (Eq a) => [a] -> [a] -> Bool
-propReplaceId xs ys =
-   replace xs xs ys == ys
-
-propReplaceCycle :: (Eq a) => [a] -> [a] -> Bool
-propReplaceCycle xs ys =
-   replace xs ys (cycle xs) == cycle ys
-
 {- | This is slightly wrong, because it re-replaces things.
      That's also the reason for inefficiency:
         The replacing can go on only when subsequent replacements are finished.
@@ -588,6 +869,9 @@ replace'' src dst =
                           then dst ++ drop (length src) y
                           else y) []
 
+{- |
+prop \src dst xs -> replace src dst xs === multiReplace [(src,dst)] (xs::String)
+-}
 multiReplace :: Eq a => [([a], [a])] -> [a] -> [a]
 multiReplace dict =
    let recourse [] = []
@@ -610,10 +894,6 @@ multiReplace' dict =
              (find (flip isPrefixOf str . fst) dict)
    in  recourse
 
-propMultiReplaceSingle :: Eq a => [a] -> [a] -> [a] -> Bool
-propMultiReplaceSingle src dst x =
-   replace src dst x == multiReplace [(src,dst)] x
-
 
 -- * Lists of lists
 
@@ -640,7 +920,7 @@ transposeFill :: [[a]] -> [[Maybe a]]
 transposeFill =
    unfoldr (\xs ->
       toMaybe (not (null xs))
-         (mapSnd (dropWhileRev null) $ unzipCons xs))
+         (mapSnd (Rev.dropWhile null) $ unzipCons xs))
 
 unzipCons :: [[a]] -> ([Maybe a], [[a]])
 unzipCons =
@@ -673,7 +953,7 @@ shear' xs@(_:_) =
        zipConc (a:as) (b:bs) = (a++b) : zipConc as bs
        zipConc [] bs = bs
        zipConc as [] = as
-   in  y : zipConc ys (shear' (dropWhileRev null zs))
+   in  y : zipConc ys (shear' (Rev.dropWhile null zs))
               {- Dropping trailing empty lists is necessary,
                  otherwise finite lists are filled with empty lists. -}
 shear' [] = []
@@ -689,6 +969,8 @@ Transform
 It's like 'shear' but the order of elements in the sub list is reversed.
 Its implementation seems to be more efficient than that of 'shear'.
 If the order does not matter, better choose 'shearTranspose'.
+
+prop> \xs -> shearTranspose xs  ===  map reverse (shear (xs::[String]))
 -}
 shearTranspose :: [[a]] -> [[a]]
 shearTranspose =
@@ -721,9 +1003,9 @@ zipCons' xs [] = map (:[]) xs
 {- |
 Operate on each combination of elements of the first and the second list.
 In contrast to the list instance of 'Monad.liftM2'
-in holds the results in a list of lists.
-It holds
-@concat (outerProduct f xs ys)  ==  liftM2 f xs ys@
+it holds the results in a list of lists.
+
+prop> \xs ys -> let f x y = (x::Char,y::Int) in concat (outerProduct f xs ys)  ===  liftM2 f xs ys
 -}
 outerProduct :: (a -> b -> c) -> [a] -> [b] -> [[c]]
 outerProduct f xs ys = map (flip map ys . f) xs
@@ -745,15 +1027,14 @@ takeWhileMulti aps@(p:ps) axs@(x:xs) =
       then x : takeWhileMulti aps xs
       else takeWhileMulti ps axs
 
+{- |
+prop> \ys xs -> let ps = map (<=) ys in takeWhileMulti ps xs === takeWhileMulti' ps (xs::String)
+-}
 takeWhileMulti' :: [a -> Bool] -> [a] -> [a]
 takeWhileMulti' ps xs =
    concatMap fst (tail
       (scanl (flip span . snd) (undefined,xs) ps))
 
-propTakeWhileMulti :: (Eq a) => [a -> Bool] -> [a] -> Bool
-propTakeWhileMulti ps xs =
-   takeWhileMulti ps xs == takeWhileMulti' ps xs
-
 {-
 Debug.QuickCheck.quickCheck (propTakeWhileMulti [(<0), (>0), odd, even, ((0::Int)==)])
 -}
@@ -768,11 +1049,11 @@ foldl'r, foldl'rStrict, foldl'rNaive ::
    (b -> a -> b) -> b -> (c -> d -> d) -> d -> [(a,c)] -> (b,d)
 foldl'r f b0 g d0 =
 --   (\(k,d1) -> (k b0, d1)) .
-   mapFst ($b0) .
+   mapFst ($ b0) .
    foldr (\(a,c) ~(k,d) -> (\b -> k $! f b a, g c d)) (id,d0)
 
 foldl'rStrict f b0 g d0 =
-   mapFst ($b0) .
+   mapFst ($ b0) .
    foldr (\(a,c) ~(k,d) -> ((,) $! (\b -> k $! f b a)) $! g c d) (id,d0)
 
 foldl'rNaive f b g d xs =
@@ -800,12 +1081,54 @@ Debug.QuickCheck.quickCheck (\b d -> propFoldl'r (+) (b::Int) (++) (d::[Int]))
 -}
 
 
+{- |
+>>> lengthAtLeast 0 ""
+True
+>>> lengthAtLeast 3 "ab"
+False
+>>> lengthAtLeast 3 "abc"
+True
+>>> lengthAtLeast 3 $ repeat 'a'
+True
+>>> lengthAtLeast 3 $ "abc" ++ undefined
+True
+
+prop> \n xs -> lengthAtLeast n (xs::String)  ===  (length xs >= n)
+-}
 lengthAtLeast :: Int -> [a] -> Bool
 lengthAtLeast n =
    if n<=0
-     then const False
+     then const True
      else not . null . drop (n-1)
 
+{- |
+>>> lengthAtMost 0 ""
+True
+>>> lengthAtMost 3 "ab"
+True
+>>> lengthAtMost 3 "abc"
+True
+>>> lengthAtMost 3 "abcd"
+False
+>>> lengthAtMost 3 $ repeat 'a'
+False
+>>> lengthAtMost 3 $ "abcd" ++ undefined
+False
+
+prop> \n xs -> lengthAtMost n (xs::String)  ===  (length xs <= n)
+-}
+lengthAtMost :: Int -> [a] -> Bool
+lengthAtMost n =
+   if n<0
+     then const False
+     else null . drop n
+
+{- |
+prop> \n xs -> lengthAtMost0 n (xs::String)  ===  (length xs <= n)
+-}
+lengthAtMost0 :: Int -> [a] -> Bool
+lengthAtMost0 n = (n>=) . length . take (n+1)
+
 {-
 Iterate until elements start to cycle.
 This implementation is inspired by Elements of Programming
@@ -846,34 +1169,61 @@ rotate, rotate', rotate'' :: Int -> [a] -> [a]
 rotate n x =
    Match.take x (drop (mod n (length x)) (cycle x))
 
-{- | more efficient implementation of rotate' -}
+{- | more efficient implementation of rotate'
+
+prop> \n (NonEmpty xs) -> rotate n xs === rotate' n (xs::String)
+-}
 rotate' n x =
    uncurry (flip (++))
            (splitAt (mod n (length x)) x)
 
+{- |
+prop> \(NonNegative n) xs -> rotate n xs === rotate'' n (xs::String)
+-}
 rotate'' n x =
    Match.take x (drop n (cycle x))
 
-propRotate :: Eq a => Int -> [a] -> Bool
-propRotate n x =
-   rotate n x == rotate'  n x &&
-   rotate n x == rotate'' n x
-{- Debug.QuickCheck.quickCheck
-      (\n x -> n>=0 Debug.QuickCheck.==>
-          List.HT.propRotate n ((0::Int):x)) -}
-
-{-|
+{- |
 Given two lists that are ordered
 (i.e. @p x y@ holds for subsequent @x@ and @y@)
 'mergeBy' them into a list that is ordered, again.
+
+>>> mergeBy (<=) "agh" "begz"
+"abegghz"
 -}
 mergeBy :: (a -> a -> Bool) -> [a] -> [a] -> [a]
-mergeBy = Key.mergeBy
+mergeBy = Cmp.merge
 
 
+{- |
+>>> allEqual "aab"
+False
+>>> allEqual "aaa"
+True
+>>> allEqual "aa"
+True
+>>> allEqual "a"
+True
+>>> allEqual ""
+True
+-}
 allEqual :: Eq a => [a] -> Bool
 allEqual = and . mapAdjacent (==)
 
+{- |
+>>> isAscending "abc"
+True
+>>> isAscending "abb"
+True
+>>> isAscending "aba"
+False
+>>> isAscending "cba"
+False
+>>> isAscending "a"
+True
+>>> isAscending ""
+True
+-}
 isAscending :: (Ord a) => [a] -> Bool
 isAscending = and . isAscendingLazy
 
@@ -883,14 +1233,92 @@ isAscendingLazy = mapAdjacent (<=)
 {- |
 This function combines every pair of neighbour elements
 in a list with a certain function.
+
+>>> mapAdjacent (<=) ""
+[]
+>>> mapAdjacent (<=) "a"
+[]
+>>> mapAdjacent (<=) "aba"
+[True,False]
+>>> mapAdjacent (,) "abc"
+[('a','b'),('b','c')]
+
+prop> \x xs -> mapAdjacent subtract (scanl (+) x xs) === (xs::[Integer])
 -}
 mapAdjacent :: (a -> a -> b) -> [a] -> [b]
 mapAdjacent f xs = zipWith f xs (tail xs)
 
+{- |
+<http://mail.haskell.org/libraries/2016-April/026912.html>
+
+prop> \xs -> mapAdjacent (,) xs === mapAdjacentPointfree (,) (xs::String)
+-}
+mapAdjacentPointfree :: (a -> a -> b) -> [a] -> [b]
+mapAdjacentPointfree f = zipWith f <*> tail
+
+
+{- |
+>>> let f x y z = [x,y]++show(z::Int) in mapAdjacent1 f 'a' [('b',1), ('c',2), ('d',3)]
+["ab1","bc2","cd3"]
+-}
+mapAdjacent1 :: (a -> a -> b -> c) -> a -> [(a,b)] -> [c]
+mapAdjacent1 f a xs =
+   zipWith (\a0 (a1,b) -> f a0 a1 b) (a : map fst xs) xs
+
+
+{- |
+>>> equalWith (<=) "ab" "bb"
+True
+>>> equalWith (<=) "aa" "bbb"
+False
+>>> equalWith (==) "aa" "aaa"
+False
+
+prop> \as bs -> let f a b = abs (a-b) <= (10::Int) in equalWith f as bs ===  equalWithRec f as bs
+prop> \as bs -> let f a b = abs (a-b) <= (10::Int) in equalWith f as bs ===  equalWithLiftM f as bs
+-}
+equalWith, equalWithLiftM, equalWithRec ::
+   (a -> b -> Bool) -> [a] -> [b] -> Bool
+equalWith f as bs =
+   and $
+   zipWith
+      (\ma mb ->
+         case (ma,mb) of
+            (Just a, Just b) -> f a b
+            (Nothing, Nothing) -> True
+            _ -> False)
+      (map Just as ++ [Nothing])
+      (map Just bs ++ [Nothing])
+
+equalWithLiftM f as bs =
+   all (Just True ==) $
+   zipWith
+      (\ma mb ->
+         case (ma,mb) of
+            (Nothing, Nothing) -> Just True
+            _ -> liftM2 f ma mb)
+      (map Just as ++ [Nothing])
+      (map Just bs ++ [Nothing])
+
+equalWithRec f =
+   let go (a:as) (b:bs) = f a b && go as bs
+       go [] [] = True
+       go _ _ = False
+   in go
+
 
 {- |
 Enumerate without Enum context.
 For Enum equivalent to enumFrom.
+
+>>> range 0 :: [Integer]
+[]
+>>> range 1 :: [Integer]
+[0]
+>>> range 8 :: [Integer]
+[0,1,2,3,4,5,6,7]
+
+prop> \(NonNegative n) -> length (range n :: [Integer]) === n
 -}
 range :: Num a => Int -> [a]
 range n = take n (iterate (+1) 0)
@@ -916,6 +1344,8 @@ The idea is:
 From the list @map (powerAssociative op a a) [0,(2*n)..]@
 we compute the list @map (powerAssociative op a a) [0,n..]@,
 and iterate that until @n==1@.
+
+prop> \x -> equating (take 1000) (List.iterate (x+) x) (iterateAssociative (+) (x::Integer))
 -}
 iterateAssociative :: (a -> a -> a) -> a -> [a]
 iterateAssociative op a =
@@ -933,6 +1363,8 @@ This also preserves log n applications per value.
 However it has a space leak,
 because for the value with index @n@
 all elements starting at @div n 2@ must be kept.
+
+prop> \x -> equating (take 1000) (List.iterate (x+) x) (iterateLeaky (+) (x::Integer))
 -}
 iterateLeaky :: (a -> a -> a) -> a -> [a]
 iterateLeaky op x =
